#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wshadow"
#include "G4RunManager.hh"
#include "G4UImanager.hh"

#include "G4UIterminal.hh"
#include "G4UItcsh.hh"

#include "DetectorConstruction.hh"
#include "PhysicsList.hh"
#include "PrimaryGeneratorAction.hh"
#include "PrimaryGeneratorAction_Messenger.hh"
#include "TrackingAction.hh"
#include "SteppingAction.hh"
#include "StackingAction.hh"
#include "EventAction.hh"
#include "EventAction_Messenger.hh"
#include "RunAction.hh"
#include "Incoming_Beam.hh"
#include "Incoming_Beam_Messenger.hh"
#include "Outgoing_Beam.hh"
#include "Outgoing_Beam_Messenger.hh"

#ifdef G4VIS_USE
#include "VisManager.hh"
#endif

#include "G4Timer.hh"
G4Timer Timer;
G4Timer Timerintern;

int main(int argc,char** argv) 
{
  // choose the Random engine
  HepRandom::setTheEngine(new RanecuEngine);
  G4long seed;
  seed = time(0);
  HepRandom::setTheSeed(seed);
  // Construct the default run manager
  G4RunManager* runManager = new G4RunManager;

  cout << "Instantiating DetectorConstruction ..." << endl;
  // set mandatory initialization classes
  DetectorConstruction* detector = new DetectorConstruction();
  runManager->SetUserInitialization(detector);

  PhysicsList *physicsList = new PhysicsList(detector);
  runManager->SetUserInitialization(physicsList);

  cout << "... Done" << endl;

  // Construct incoming and outgoing beams
  cout << "Construct incoming and outgoing beams" << endl;
  Incoming_Beam* BeamIn = new Incoming_Beam();
  Incoming_Beam_Messenger* IncomingBeamMessenger = new Incoming_Beam_Messenger(BeamIn);

  Outgoing_Beam* BeamOut = new Outgoing_Beam();
  BeamOut->defaultIncomingIon(BeamIn);
  physicsList->SetOutgoingBeam(BeamOut);
  Outgoing_Beam_Messenger* OutgoingBeamMessenger = new Outgoing_Beam_Messenger(BeamOut);
  cout << "... Done" << endl;

  // set mandatory user action class
  cout << "Event Action" << endl;
  EventAction* eventAction = new EventAction();
  EventAction_Messenger* eventActionMessenger = new EventAction_Messenger(eventAction);
  runManager->SetUserAction(eventAction);
  cout << "... Done" << endl;

  cout << "Primary Generator" << endl;
  PrimaryGeneratorAction* generatorAction = new PrimaryGeneratorAction(detector,BeamIn,BeamOut);
  PrimaryGeneratorAction_Messenger* generatorActionMessenger = new PrimaryGeneratorAction_Messenger(generatorAction);
  cout << "... Done" << endl;

  cout << "Run Action" << endl;
  runManager->SetUserAction(generatorAction);
  RunAction* runAction = new RunAction(detector,BeamOut,eventAction);
  runManager->SetUserAction(runAction);
  cout << "... Done" << endl;

  cout << "Tracking Action" << endl;
  TrackingAction* trackingAction = new TrackingAction(eventAction);
  runManager->SetUserAction(trackingAction);
  cout << "... Done" << endl;

  cout << "Stepping Action" << endl;
  SteppingAction* steppingAction = new SteppingAction();
  runManager->SetUserAction(steppingAction);
  cout << "... Done" << endl;

  cout << "Stacking Action" << endl;
  StackingAction* stacking_action = new StackingAction();
  runManager->SetUserAction(stacking_action);
  cout << "... Done" << endl;

  G4UIsession* session=0;

#ifdef G4VIS_USE
  // visualization manager
  cout << "Starting visualization...";
  G4VisManager* visManager=0;
#endif

  if (argc==1)   // Define UI session for interactive mode.
    {

#ifdef G4VIS_USE
      // visualization manager
      visManager = new VisManager; 
      visManager->Initialize();
      cout << "Done!" << endl;
#endif

// G4UIterminal is a (dumb) terminal.
#ifdef G4UI_USE_TCSH
      session = new G4UIterminal(new G4UItcsh);      
#else
      session = new G4UIterminal();
#endif

    }

  // Initialize G4 kernel
  cout << "*** Initializing runManager" << endl;
  runManager->SetVerboseLevel(0);
  runManager->Initialize();
  cout << "*** Initialized runManager" << endl;

  // get the pointer to the UI manager and set verbosities
  G4UImanager* UI = G4UImanager::GetUIpointer();

  if (session)   // Define UI session for interactive mode.
    {
      session->SessionStart();
      delete session;
    }
  else           // Batch mode
    { 
      G4String command = "/control/execute ";
      G4String fileName = argv[1];
      UI->ApplyCommand(command+fileName);
    }

  // job termination
  if(argc==1){
#ifdef G4VIS_USE
  delete visManager;
#endif
  }
  delete runManager;
  delete BeamIn;
  delete IncomingBeamMessenger;
  delete BeamOut;
  delete OutgoingBeamMessenger;
  delete eventActionMessenger;
  delete generatorActionMessenger;

  return 0;
}
#pragma GCC diagnostic pop
