#include "DetectorConstruction.hh"

DetectorConstruction::DetectorConstruction()
{

  targetStatus   = false;
  beamTubeStatus = false;

  gretinaStatus = true;

  myMessenger = new DetectorConstruction_Messenger(this);
}

DetectorConstruction::~DetectorConstruction()
{
  delete ExperimentalHallMessenger;
  delete TargetMessenger;
  delete TrackerIonSDMessenger;
  delete TrackerGammaSDMessenger;
  delete BeamTubeMessenger;
}

G4VPhysicalVolume* DetectorConstruction::Construct()
{

  //  DefineMaterials();
  materials = new Materials();

  // Tracker for ions in the target

  G4SDManager* SDman = G4SDManager::GetSDMpointer();
  TrackerIon = new TrackerIonSD("IonTracker");
  TrackerIonSDMessenger = new TrackerIonSD_Messenger(TrackerIon);
  SDman->AddNewDetector( TrackerIon );

  // Tracker for gammas in GRETINA

  TrackerGamma = new TrackerGammaSD("GammaTracker");
  TrackerGammaSDMessenger = new TrackerGammaSD_Messenger(TrackerGamma);
  SDman->AddNewDetector( TrackerGamma );

  // Experimental Hall

  Experimental_Hall* ExperimentalHall = new Experimental_Hall(materials);
  ExpHall_phys = ExperimentalHall->Construct();
  ExpHall_log  = ExperimentalHall->GetLogVolume();
  ExperimentalHallMessenger = new Experimental_Hall_Messenger(ExperimentalHall);

  G4RunManager* runManager = G4RunManager::GetRunManager();
  runManager->DefineWorldVolume( ExpHall_phys );

  // Background sphere

  BackgroundSphere = new Background_Sphere(ExpHall_log,materials);
  BackgroundSphereMessenger = new Background_Sphere_Messenger(BackgroundSphere);
  BackgroundSphere->Construct();

  // Beam Tube

  BeamTube = new Beam_Tube(ExpHall_log,materials);
  BeamTubeMessenger = new Beam_Tube_Messenger(BeamTube);

  // Target

  aTarget = new Target(ExpHall_log,materials);
  TargetMessenger = new Target_Messenger(aTarget);

  // GRETINA

  the_Gretina_Array = new Gretina_Array();
  the_Gretina_Array_Messenger = new Gretina_Array_Messenger(the_Gretina_Array);

  return ExpHall_phys;
}

void DetectorConstruction::DefineMaterials()
{
  // Remove this method. All materials are now defined by the Materials class.
}

void DetectorConstruction::Placement()
{

  // Beam Tube
  if( beamTubeStatus ){
    BeamTube->Construct();
    BeamTube->ConstructShield(0);
    BeamTube->ConstructShield(1);
  }

  // Target
  if( targetStatus ){
    aTarget->Construct();
    aTarget->GetTargetLog()->SetSensitiveDetector(TrackerIon);
  }

  if(gretinaStatus)
    the_Gretina_Array->Placement();

}

///////////////////
// The Messenger
///////////////////

DetectorConstruction_Messenger::DetectorConstruction_Messenger(DetectorConstruction* pTarget):myTarget(pTarget)
{

  const char *aLine;
  G4String commandName;

  commandName = "/HRarray/Construct";
  aLine = commandName.c_str();
  UpdateCmd = new G4UIcmdWithoutParameter(aLine, this);
  UpdateCmd->SetGuidance("Update geometry.");
  UpdateCmd->AvailableForStates(G4State_PreInit,G4State_Idle);

  commandName = "/Target/Construct";
  aLine = commandName.c_str();
  TargetCmd = new G4UIcmdWithoutParameter(aLine, this);
  TargetCmd->SetGuidance("Construct the target.");
  TargetCmd->AvailableForStates(G4State_PreInit,G4State_Idle);

  commandName = "/HRarray/NoDetectors";
  aLine = commandName.c_str();
  NoGretCmd = new G4UIcmdWithoutParameter(aLine, this);
  NoGretCmd->SetGuidance("Omit the GRETINA detectors.");
  NoGretCmd->AvailableForStates(G4State_PreInit,G4State_Idle);

  commandName = "/BeamTube/Construct";
  aLine = commandName.c_str();
  BeamTubeCmd = new G4UIcmdWithoutParameter(aLine, this);
  BeamTubeCmd->SetGuidance("Construct the beam tube.");
  BeamTubeCmd->AvailableForStates(G4State_PreInit,G4State_Idle);

}

DetectorConstruction_Messenger::~DetectorConstruction_Messenger()
{
  delete UpdateCmd;
  delete TargetCmd;
  delete NoGretCmd;
  delete BeamTubeCmd;
}

void DetectorConstruction_Messenger::SetNewValue(G4UIcommand* command,G4String newValue)
{ 
  if( command == UpdateCmd ) {
    myTarget->Placement();
  } 
  if( command == TargetCmd ) {
    myTarget->SetTargetStatus(true);
  } 
  if( command == NoGretCmd ) {
    myTarget->SetGretinaStatus(false);
  }
  if( command == BeamTubeCmd ) {
    myTarget->SetBeamTubeStatus(true);
  } 
  if(newValue)
    G4cout <<"newValue " << newValue << G4endl;
}
